# STL
import re
from abc import ABC, abstractmethod
from typing import Set
from functools import lru_cache as cache  # cache comes in 3.9

# PDM
import regex
from typing_extensions import override

# LOCAL
from sonatoki.constants import (
    VOWELS,
    CONSONANTS,
    NIMI_PU_SET,
    ALPHABET_SET,
    UNICODE_PUNCT,
    ALLOWABLES_SET,
    NIMI_LINKU_SET,
    NIMI_PU_ALE_SET,
    NIMI_LINKU_ALE_SET,
    PRUNED_POSIX_PUNCT,
    NIMI_LINKU_SANDBOX_SET,
)

regex.DEFAULT_VERSION = regex.VERSION1


class Filter(ABC):
    @classmethod
    @abstractmethod
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        raise NotImplementedError


class RegexFilter(Filter):
    pattern: "re.Pattern[str]"

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return not not re.fullmatch(cls.pattern, token)


class Regex1Filter(Filter):
    pattern: "regex.Pattern[str]"

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return not not regex.fullmatch(cls.pattern, token)


class SetFilter(Filter):
    tokens: Set[str]

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return token.lower() in cls.tokens


class Miscellaneous(SetFilter):
    tokens = ALLOWABLES_SET


class ProperName(Filter):
    """Determines if a given token is a valid name (also called a loan word).
    When Toki Pona is written with the Latin alphabet, names are generally
    capitalized at their start. This filter identifies those tokens.

    Note that this alone cannot determine if a token is a valid name, because
    a standalone name is considered invalid in Toki Pona- names generally have head nouns.
    This tool only examines one token at a time, so cannot detect names any better than identifying their capital letter.
    """

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        return token == token.capitalize()


class NimiPu(SetFilter):
    tokens = NIMI_PU_SET


class NimiPuAle(SetFilter):
    tokens = NIMI_PU_ALE_SET


class NimiLinku(SetFilter):
    tokens = NIMI_LINKU_SET


class NimiLinkuAle(SetFilter):
    tokens = NIMI_LINKU_ALE_SET


class NimiLinkuSandbox(SetFilter):
    tokens = NIMI_LINKU_SANDBOX_SET


class Phonotactic(RegexFilter):
    """Determines if a given token is phonotactically valid Toki Pona (or `n`).
    Excludes both consecutive nasals and the illegal syllables:
    - "nm", "nn"
    - "wu", "wo", "ji", "ti"

    Note that if this validator is used after `Cleaners.ConsecutiveDuplicates`,
    "nn" cannot be found."""

    pattern = re.compile(
        rf"^((^[{VOWELS}]|[klmnps][{VOWELS}]|[jt][aeou]|[w][aei])(n(?![mn]))?)+$|^n$",
        # Can't split initial vowel group off like in Syllabics because of
        # consecutive nasal detection; it is costly to duplicate
        flags=re.IGNORECASE,
    )


class Syllabic(RegexFilter):
    """Determines if a given token is syllabically valid Toki Pona (or `n`).
    Words must have correctly ordered vowels and consonants, but the phonotactic
    exceptions are not considered."""

    # rf"^((^[{VOWELS}]|[{CONSONANTS}][{VOWELS}])n?)+$|^n$"
    # Alterative I was exploring takes ~15% more steps
    pattern = re.compile(
        rf"^(?:^[{VOWELS}]n?)?(?:[{CONSONANTS}][{VOWELS}]n?)*$|^n$",
        flags=re.IGNORECASE,
    )


class Alphabetic(Filter):
    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, token: str) -> bool:
        # Faster than regex version
        return set(token.lower()).issubset(ALPHABET_SET)


class Numeric(Filter):
    """Determine if a given token is entirely numeric.
    Covers all numeric symbols in Unicode.

    This will fail to find numeric tokens such as "1.111" or "-42",
    but if used with the aggressive tokenizer designed for `tok`, these will be
    split into `["1", ".", "111"]` and `["-", "42"]` respectively. As such, the
    numeric tokens will be split from their punctuation."""

    @classmethod
    @override
    @cache(maxsize=None)
    def filter(cls, msg: str) -> bool:
        return msg.isnumeric()


class Punctuation(RegexFilter):
    pattern = re.compile(rf"[{PRUNED_POSIX_PUNCT}{UNICODE_PUNCT}]+")


__all__ = [
    "NimiPu",
    "NimiLinku",
    "NimiLinkuAle",
    "Phonotactic",
    "Syllabic",
    "Alphabetic",
    "ProperName",
    "Punctuation",
    "Numeric",
]



# STL
import string

# PDM
import regex as re
import hypothesis.strategies as st
from hypothesis import HealthCheck, given, assume, example, settings

# LOCAL
from sonatoki.Filters import (
    NimiPu,
    Numeric,
    Syllabic,
    NimiLinku,
    Alphabetic,
    ProperName,
    Phonotactic,
    Punctuation,
)
from sonatoki.Cleaners import ConsecutiveDuplicates
from sonatoki.constants import NIMI_PU, NIMI_LINKU

# FILESYSTEM
from .test_utils import ALPHABETIC_RE, PROPER_NAME_RE


@given(st.sampled_from(NIMI_PU))
@example("lukin")
@example("selo")
@example("li")
def test_NimiPu(s: str):
    res = NimiPu.filter(s)
    assert res, repr(s)


@given(st.sampled_from(NIMI_LINKU))
@example("pona")
@example("tonsi")
@example("kipisi")
@example("n")
def test_NimiLinku(s: str):
    res = NimiLinku.filter(s)
    assert res, repr(s)


@given(st.sampled_from(NIMI_LINKU))
def test_nimi_linku_properties(s: str):
    assert ConsecutiveDuplicates.clean(s) == s, repr(s)
    assert Alphabetic.filter(s), repr(s)
    assert Syllabic.filter(s), repr(s)
    assert Phonotactic.filter(s), repr(s)
    # Passing phonotactic implies all of the above


@given(st.from_regex(Phonotactic.pattern.pattern, fullmatch=True))
@example("kijetesantakalu")
@example("n")
def test_Phonotactic(s: str):
    res = Phonotactic.filter(s)
    assert res, repr(s)


@given(st.from_regex(Syllabic.pattern.pattern, fullmatch=True))
@example("wuwojitiwunwonjintinmanna")
def test_Syllabic(s: str):
    res = Syllabic.filter(s)
    assert res, repr(s)


@given(st.from_regex(ALPHABETIC_RE, fullmatch=True))
@example("muems")
@example("mpptp")
@example("tptpt")
def test_Alphabetic(s: str):
    res = Alphabetic.filter(s)
    assert res, repr(s)


@given(st.from_regex(PROPER_NAME_RE, fullmatch=True))
def test_ProperName(s: str):
    res = ProperName.filter(s)
    assert res, repr(s)


@given(st.from_regex(Punctuation.pattern.pattern, fullmatch=True))
@example("[]")
@example(r"\\")
@example(r"\"")
@example("⟨·⟩")
@example("…")
@example("「」")  # `　`
@example(string.punctuation)
def test_Punctuation(s: str):
    res = Punctuation.filter(s)
    assert res, repr(s)


@given(st.from_regex(r"\d+", fullmatch=True))
@example("124125")
@example("99990000")
def test_Numeric(s: str):
    res = Numeric.filter(s)
    assert res, repr(s)



# STL
import itertools

# PDM
import hypothesis.strategies as st

# LOCAL
from sonatoki.Filters import Syllabic, Phonotactic
from sonatoki.constants import ALPHABET, NIMI_LINKU_ALE

PROPER_NAME_RE = r"[A-Z][a-z]*"
ALPHABETIC_RE = rf"[{ALPHABET}{ALPHABET.upper()}]+"

token_strategy = (
    st.sampled_from(NIMI_LINKU_ALE)
    | st.from_regex(Phonotactic.pattern.pattern, fullmatch=True)
    | st.from_regex(Syllabic.pattern.pattern, fullmatch=True)
    | st.from_regex(PROPER_NAME_RE, fullmatch=True)
    | st.from_regex(ALPHABETIC_RE, fullmatch=True)
)


token_list_strategy = st.lists(
    token_strategy,
    min_size=0,
    max_size=10,
    unique=True,
)


def overlapping_pairs(iterable: str):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    _ = next(b, None)
    return zip(a, b)



# PDM
import pytest

# LOCAL
from sonatoki.ilo import Ilo
from sonatoki.Configs import LazyConfig, PrefConfig


@pytest.fixture
def ilo():
    ilo = Ilo(**PrefConfig)
    return ilo


@pytest.fixture()
def lazy_ilo():
    ilo = Ilo(**LazyConfig)
    # ilo.logging_threshold = 0.8
    return ilo


ALL_VALID = [
    "mi unpa e mama sina",
    "mama sina li lon seme? mi wile toki tawa ona",
    "sina sike pakala",
    "    sina    seme     e     mi     ?",
    "AAAAAAAAAAA",
    "muuuu MUUU muUuUuU",
    "wawa mute. " * 10,
]

IGNORABLES = [
    "",
    " ",
    "2+2=5",
    "kiwen moli 42",
    "https://mun.la/sona",
    "https://example.com/",
    "mi wile e ni: <https://example.com> li pona",
    "lipu https://example.com li kama pona",
    "...",
    " ⟨·⟩, a",
    "·····",
]

SYLLABIC_MATCHES = [
    "ni li tenpo penpo",
    "sipisi",
    "walawa malama walama malama mupi",
    "mi sona ala e nimi sunopatikuna",
    "kalama wuwojiti li pana e sona",
    "jan Awaja en jan Alasali en jan Akesinu li pona",  # syllables match before names here
]

ALPHABETIC_MATCHES = [
    "mi mtue o kama sona",
    "mi mute o kma son",
    "ni li tptpt",
    "mi wile pana lon sptp",
    "tmo tawa mi li pona mute la mi kepeken ona lon tenpo mute",
    "mi pakla lon nimi pi mute lili, taso ale li pona tan ni: mi toki mute",
]

NAME_MATCHES = [
    "musi Homestuck li ike tawa mi",
    "ilo Google li sona ala e nimi Emoticon la mi wile utala e ona",
    "toki Kanse li lon",
    "toki Lojban li nasa e lawa mi",
]

SOME_INVALID = ["kulupu xerox li ike", "mi tawa ma ohio"]


EXCESSIVE_SYLLABICS = [
    "manama manama namana namana majani makala",
]

EXCESSIVE_ALPHABETICS = [
    "21st",  # candidate for xfails?
    "tok",
    "mut",
    "mtue",
    "I wait, I sulk, as a tool I make stoops to ineptness.",
    "aaa i non-saw usa's most multiple element-set. it's as asinine as in `e`-less speak",
    "mi pakla ln tepo mtue ls mi kn ala tok poan aun seem",
    "so, to atone like papa—an awesome anon (no-name) sin man—i ate an asinine lemon-limelike tomato jalapeno isotope. 'nonsense!' amen. note to Oman: take mine katana to imitate a ninja in pantomime. atomise one nuke? 'insane misuse!' same. likewise, Susan, awaken a pepino melon in a linen pipeline. (penile) emanate semen. joke: manipulate a tame toneme to elope online tonite",
]

EXCESSIVE_NAMES = [
    "I Want To Evade The Filter",
    "If You Do This The Bot Can't See You",
    "This Is A Statement In Perfect Toki Pona, I Guarantee",
]

NON_MATCHES = [
    "bong",
    "super bruh moment 64",
    "homestuck",
    "homestuck Homestuck",
]

XFAILS = [
    "lete li ike x.x",  # emoticon should not be a problem
]


@pytest.mark.parametrize(
    "text",
    ALL_VALID
    + SYLLABIC_MATCHES
    + ALPHABETIC_MATCHES
    + NAME_MATCHES
    + SOME_INVALID
    + IGNORABLES,
)
def test_known_good(ilo: Ilo, lazy_ilo: Ilo, text: str):
    assert ilo.is_toki_pona(text), text


@pytest.mark.parametrize(
    "text", EXCESSIVE_SYLLABICS + EXCESSIVE_ALPHABETICS + EXCESSIVE_NAMES + NON_MATCHES
)
def test_known_bad(ilo: Ilo, text: str):
    assert not ilo.is_toki_pona(text), text


@pytest.mark.parametrize(
    "text",
    ALL_VALID
    + SYLLABIC_MATCHES
    + ALPHABETIC_MATCHES
    + NAME_MATCHES
    + SOME_INVALID
    + IGNORABLES,
)
def test_known_good_lazy(lazy_ilo: Ilo, text: str):
    assert lazy_ilo.is_toki_pona(text), text
    # assumption: lazy ilo should pass anything the more strict ilo does


@pytest.mark.parametrize("text", NON_MATCHES)
def test_known_bad_lazy(lazy_ilo: Ilo, text: str):
    assert not lazy_ilo.is_toki_pona(text), text


@pytest.mark.parametrize(
    "text", EXCESSIVE_SYLLABICS + EXCESSIVE_ALPHABETICS + EXCESSIVE_NAMES
)
def test_weakness_of_lazy(lazy_ilo: Ilo, text: str):
    # NOTE: This is demonstrative, not preferential
    assert lazy_ilo.is_toki_pona(text), text


@pytest.mark.xfail
@pytest.mark.parametrize("text", XFAILS)
def test_known_xfails(ilo: Ilo, text: str):
    assert ilo.is_toki_pona(text)



# STL
from copy import deepcopy
from typing import List, Type, TypedDict

# PDM
from typing_extensions import NotRequired

# LOCAL
from sonatoki.Filters import (
    Filter,
    NimiPu,
    Numeric,
    Syllabic,
    NimiLinku,
    NimiPuAle,
    Alphabetic,
    ProperName,
    Phonotactic,
    Punctuation,
    NimiLinkuAle,
)
from sonatoki.Scorers import Number, Scorer, PassFail, SoftScaling, SoftPassFail
from sonatoki.Cleaners import Cleaner, ConsecutiveDuplicates
from sonatoki.Tokenizers import Tokenizer, WordTokenizerTok
from sonatoki.Preprocessors import (
    URLs,
    Preprocessor,
    DiscordEmotes,
    DiscordSpecial,
    DiscordChannels,
    DiscordMentions,
)


class IloConfig(TypedDict):
    preprocessors: List[Type[Preprocessor]]
    word_tokenizer: Type[Tokenizer]
    cleaners: List[Type[Cleaner]]
    ignoring_filters: List[Type[Filter]]
    scoring_filters: List[Type[Filter]]
    scorer: Type[Scorer]
    passing_score: Number


BaseConfig: IloConfig = {
    "preprocessors": [URLs],
    "cleaners": [ConsecutiveDuplicates],
    "ignoring_filters": [Numeric, Punctuation],
    "scoring_filters": [],
    "scorer": PassFail,
    "passing_score": 0.8,
    "word_tokenizer": WordTokenizerTok,
}


PrefConfig: IloConfig = deepcopy(BaseConfig)
PrefConfig["scoring_filters"].extend([NimiLinku, Syllabic, ProperName, Alphabetic])
PrefConfig["scorer"] = SoftScaling


LazyConfig: IloConfig = deepcopy(BaseConfig)
LazyConfig["scoring_filters"].extend([Alphabetic, ProperName])
LazyConfig["scorer"] = SoftPassFail

DiscordConfig: IloConfig = deepcopy(PrefConfig)
DiscordConfig["preprocessors"].extend(
    [DiscordEmotes, DiscordMentions, DiscordChannels, DiscordSpecial]
)
TelegramConfig: IloConfig = deepcopy(PrefConfig)
ForumConfig: IloConfig = deepcopy(PrefConfig)

__all__ = [
    "IloConfig",
    "BaseConfig",
    "PrefConfig",
    "LazyConfig",
    "DiscordConfig",
    "TelegramConfig",
    "ForumConfig",
]



# STL
import math
import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Type, Union

# PDM
from typing_extensions import override

# LOCAL
from sonatoki.Filters import Filter

Number = Union[int, float]
Weights = Dict[str, Number]


def sigmoid(n: int) -> Number:
    return 1 / (1 + math.exp(-(0.30 * (n - 1))))
    # n-1 makes sigmoid(1) == 0.5
    # 0.30 softens scaling in favor of short input
    # return n / (1+abs(n))   # too weak in 0.7+


class Scorer(ABC):
    @classmethod
    @abstractmethod
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        raise NotImplementedError


class PassFail(Scorer):
    """The token passes any filter or fails all of them, scoring 1 or 0 respectively."""

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]]) -> Number:
        for f in filters:
            if f.filter(token):
                return 1
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_tokens = len(tokens)
        for token in tokens:
            total_score += cls.score_token(token, filters)
        return total_score / len_tokens if len_tokens else 0


class SoftPassFail(PassFail):
    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_tokens = len(tokens)
        for token in tokens:
            total_score += cls.score_token(token, filters)

        percentage = total_score / len_tokens if len_tokens else 0
        percentage **= sigmoid(len_tokens)
        return percentage


class Scaling(Scorer):
    """
    The sooner a token matches a filter, the higher its score.
    In other words, filter order matters, weighing earlier listed filters higher than later ones.
    This is desirable to avoid messages which would only match weaker filters, as these are less likely to be Toki Pona.
    """

    @classmethod
    def score_token(cls, token: str, filters: List[Type[Filter]], scale: int):
        for i, f in enumerate(filters):
            if f.filter(token):
                return scale - i
        return 0

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_filters = len(filters)
        max_score = len(tokens) * len_filters
        for token in tokens:
            total_score += cls.score_token(token, filters, len_filters)
        return total_score / max_score if max_score else 0


class SoftScaling(Scaling):
    """Shorter messages are subject to less harsh scoring
    by mapping the token count to [0.5, 1.0] via the sigmoid function,
    then raising the score to the resultant power.
    For example, a single token scoring 0.64 will now score 0.8.
    """

    @classmethod
    @override
    def score(cls, tokens: List[str], filters: List[Type[Filter]]) -> Number:
        if not tokens:
            return 1

        total_score = 0
        len_filters = len(filters)
        len_tokens = len(tokens)

        max_score = len_tokens * len_filters
        for token in tokens:
            total_score += cls.score_token(token, filters, len_filters)

        percentage = total_score / max_score if max_score else 0
        percentage **= sigmoid(len_tokens)
        return percentage


class Logarithmic(Scorer): ...


__all__ = ["PassFail", "SoftPassFail", "Scaling", "SoftScaling"]



# STL

# PDM
import hypothesis.strategies as st
from hypothesis import assume, given, example, reproduce_failure

# LOCAL
from sonatoki.Cleaners import ConsecutiveDuplicates

# FILESYSTEM
from .test_utils import overlapping_pairs


@given(st.from_regex(ConsecutiveDuplicates.pattern.pattern))
@example("tooooki a")
@example("muuuuuu")
@example("nnn")
@example("")
@example(
    "manna"
)  # syllabically valid, but not phonotactically valid; errantly matches phonotactic filter after this cleaner
def test_ConsecutiveDuplicates(s: str):
    _ = assume("\n" not in s)
    res = ConsecutiveDuplicates.clean(s)
    for a, b in overlapping_pairs(res):
        assert a != b, (s, res)



# STL
import json
from typing import Dict, List
from pathlib import Path

LINKU = Path(__file__).resolve().parent / Path("linku.json")
SANDBOX = Path(__file__).resolve().parent / Path("sandbox.json")

VOWELS = "aeiou"
CONSONANTS = "jklmnpstw"
ALPHABET = VOWELS + CONSONANTS
ALPHABET_SET = set(ALPHABET)

LANGUAGE = "english"  # for NLTK

# `\p{posix_punct}` character class
POSIX_PUNCT = r"""-!"#$%&'()*+,./:;<=>?@[\]^_`{|}~"""
PRUNED_POSIX_PUNCT = r"""$+<=>^`|~"""  # only those that are not in UNICODE_PUNCT

# `\p{Punctuation}` character class
UNICODE_PUNCT = r"""!"#%&'()*,-./:;?@\[\\\]_{}¡§«¶·»¿;·՚՛՜՝՞՟։֊־׀׃׆׳״؉؊،؍؛؝؞؟٪٫٬٭۔܀܁܂܃܄܅܆܇܈܉܊܋܌܍߷߸߹࠰࠱࠲࠳࠴࠵࠶࠷࠸࠹࠺࠻࠼࠽࠾࡞।॥॰৽੶૰౷಄෴๏๚๛༄༅༆༇༈༉༊་༌།༎༏༐༑༒༔༺༻༼༽྅࿐࿑࿒࿓࿔࿙࿚၊။၌၍၎၏჻፠፡።፣፤፥፦፧፨᐀᙮᚛᚜᛫᛬᛭᜵᜶។៕៖៘៙៚᠀᠁᠂᠃᠄᠅᠆᠇᠈᠉᠊᥄᥅᨞᨟᪠᪡᪢᪣᪤᪥᪦᪨᪩᪪᪫᪬᪭᭚᭛᭜᭝᭞᭟᭠᭽᭾᯼᯽᯾᯿᰻᰼᰽᰾᰿᱾᱿᳀᳁᳂᳃᳄᳅᳆᳇᳓‐‑‒–—―‖‗‘’‚‛“”„‟†‡•‣․‥…‧‰‱′″‴‵‶‷‸‹›※‼‽‾‿⁀⁁⁂⁃⁅⁆⁇⁈⁉⁊⁋⁌⁍⁎⁏⁐⁑⁓⁔⁕⁖⁗⁘⁙⁚⁛⁜⁝⁞⁽⁾₍₎⌈⌉⌊⌋〈〉❨❩❪❫❬❭❮❯❰❱❲❳❴❵⟅⟆⟦⟧⟨⟩⟪⟫⟬⟭⟮⟯⦃⦄⦅⦆⦇⦈⦉⦊⦋⦌⦍⦎⦏⦐⦑⦒⦓⦔⦕⦖⦗⦘⧘⧙⧚⧛⧼⧽⳹⳺⳻⳼⳾⳿⵰⸀⸁⸂⸃⸄⸅⸆⸇⸈⸉⸊⸋⸌⸍⸎⸏⸐⸑⸒⸓⸔⸕⸖⸗⸘⸙⸚⸛⸜⸝⸞⸟⸠⸡⸢⸣⸤⸥⸦⸧⸨⸩⸪⸫⸬⸭⸮⸰⸱⸲⸳⸴⸵⸶⸷⸸⸹⸺⸻⸼⸽⸾⸿⹀⹁⹂⹃⹄⹅⹆⹇⹈⹉⹊⹋⹌⹍⹎⹏⹒⹓⹔⹕⹖⹗⹘⹙⹚⹛⹜⹝、。〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〽゠・꓾꓿꘍꘎꘏꙳꙾꛲꛳꛴꛵꛶꛷꡴꡵꡶꡷꣎꣏꣸꣹꣺꣼꤮꤯꥟꧁꧂꧃꧄꧅꧆꧇꧈꧉꧊꧋꧌꧍꧞꧟꩜꩝꩞꩟꫞꫟꫰꫱꯫﴾﴿︐︑︒︓︔︕︖︗︘︙︰︱︲︳︴︵︶︷︸︹︺︻︼︽︾︿﹀﹁﹂﹃﹄﹅﹆﹇﹈﹉﹊﹋﹌﹍﹎﹏﹐﹑﹒﹔﹕﹖﹗﹘﹙﹚﹛﹜﹝﹞﹟﹠﹡﹣﹨﹪﹫！＂＃％＆＇（）＊，－．／：；？＠［＼］＿｛｝｟｠｡｢｣､･𐄀𐄁𐄂𐎟𐏐𐕯𐡗𐤟𐤿𐩐𐩑𐩒𐩓𐩔𐩕𐩖𐩗𐩘𐩿𐫰𐫱𐫲𐫳𐫴𐫵𐫶𐬹𐬺𐬻𐬼𐬽𐬾𐬿𐮙𐮚𐮛𐮜𐺭𐽕𐽖𐽗𐽘𐽙𐾆𐾇𐾈𐾉𑁇𑁈𑁉𑁊𑁋𑁌𑁍𑂻𑂼𑂾𑂿𑃀𑃁𑅀𑅁𑅂𑅃𑅴𑅵𑇅𑇆𑇇𑇈𑇍𑇛𑇝𑇞𑇟𑈸𑈹𑈺𑈻𑈼𑈽𑊩𑑋𑑌𑑍𑑎𑑏𑑚𑑛𑑝𑓆𑗁𑗂𑗃𑗄𑗅𑗆𑗇𑗈𑗉𑗊𑗋𑗌𑗍𑗎𑗏𑗐𑗑𑗒𑗓𑗔𑗕𑗖𑗗𑙁𑙂𑙃𑙠𑙡𑙢𑙣𑙤𑙥𑙦𑙧𑙨𑙩𑙪𑙫𑙬𑚹𑜼𑜽𑜾𑠻𑥄𑥅𑥆𑧢𑨿𑩀𑩁𑩂𑩃𑩄𑩅𑩆𑪚𑪛𑪜𑪞𑪟𑪠𑪡𑪢𑬀𑬁𑬂𑬃𑬄𑬅𑬆𑬇𑬈𑬉𑱁𑱂𑱃𑱄𑱅𑱰𑱱𑻷𑻸𑽃𑽄𑽅𑽆𑽇𑽈𑽉𑽊𑽋𑽌𑽍𑽎𑽏𑿿𒑰𒑱𒑲𒑳𒑴𒿱𒿲𖩮𖩯𖫵𖬷𖬸𖬹𖬺𖬻𖭄𖺗𖺘𖺙𖺚𖿢𛲟𝪇𝪈𝪉𝪊𝪋𞥞𞥟"""
# NOTE: This list diverges slightly from the raw list, since []\ must be escaped
# The [] need to be escaped to avoid prematurely closing the regex character class
# The \ needs to be escaped to be considered as a raw \

# https://www.compart.com/en/unicode/category
# https://unicode.org/Public/UNIDATA/UnicodeData.txt


"""Commonly occurring strings which are some kind of valid Toki Pona or external token"""
ALLOWABLES = {
    "cw",  # Content Warning
    "x",  # ala
    "y",  # anu
    "kxk",  # ken ala ken
    "wxw",  # wile ala wile
}


with open(LINKU) as f:
    r: Dict[str, Dict[str, str]] = json.loads(f.read())
    NIMI_PU: List[str] = [d["word"] for d in r.values() if d["book"] == "pu"]
    NIMI_PU_ALE: List[str] = NIMI_PU + ["namako", "kin", "oko"]
    NIMI_LINKU: List[str] = [
        d["word"] for d in r.values() if d["usage_category"] in ["core", "common"]
    ]
    NIMI_LINKU_ALE: List[str] = [d["word"] for d in r.values()]

with open(SANDBOX) as f:
    r: Dict[str, Dict[str, str]] = json.loads(f.read())
    NIMI_LINKU_SANDBOX: List[str] = [d["word"] for d in r.values()]


NIMI_PU_SET = set(NIMI_PU)
NIMI_PU_ALE_SET = set(NIMI_PU_ALE)
NIMI_LINKU_SET = set(NIMI_LINKU)
NIMI_LINKU_ALE_SET = set(NIMI_LINKU_ALE)
NIMI_LINKU_SANDBOX_SET = set(NIMI_LINKU_SANDBOX)
ALLOWABLES_SET = set(ALLOWABLES)

__all__ = [
    "VOWELS",
    #
    "CONSONANTS",
    #
    "ALPHABET",
    "ALPHABET_SET",
    #
    "NIMI_PU",
    "NIMI_PU_SET",
    #
    "NIMI_PU_ALE",
    "NIMI_PU_ALE_SET",
    #
    "NIMI_LINKU",
    "NIMI_LINKU_SET",
    #
    "NIMI_LINKU_ALE",
    "NIMI_LINKU_ALE_SET",
    #
    "NIMI_LINKU_SANDBOX",
    "NIMI_LINKU_SANDBOX_SET",
]



# STL
import re
from abc import ABC, abstractmethod
from typing import List

# PDM
import regex
from typing_extensions import override

# LOCAL
from sonatoki.constants import UNICODE_PUNCT, PRUNED_POSIX_PUNCT

try:
    # PDM
    import nltk
    from nltk.tokenize import sent_tokenize as __sent_tokenize_nltk
    from nltk.tokenize import word_tokenize as __word_tokenize_nltk
except ImportError as e:
    nltk = e


regex.DEFAULT_VERSION = regex.VERSION1


class Tokenizer(ABC):
    @classmethod
    @abstractmethod
    def tokenize(cls, s: str) -> List[str]: ...


class NoOpTokenizer(Tokenizer):
    """This is a special case that you do not want or need."""

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [s]


class RegexTokenizer(Tokenizer):
    pattern: "re.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [clean for word in re.split(cls.pattern, s) if (clean := word.strip())]


class Regex1Tokenizer(Tokenizer):
    pattern: "regex.Pattern[str]"

    @classmethod
    @override
    def tokenize(cls, s: str) -> List[str]:
        return [
            clean for word in regex.split(cls.pattern, s) if (clean := word.strip())
        ]


class WordTokenizerTok(RegexTokenizer):
    pattern = re.compile(rf"""([{PRUNED_POSIX_PUNCT}{UNICODE_PUNCT}]+|\s+)""")


class SentTokenizerTok(RegexTokenizer):
    pattern = re.compile(r"""(?<=[.?!:;·…“”"'()\[\]\-])|$""", flags=re.MULTILINE)
    # TODO: are <> or {} that common as *sentence* delims? [] are already a stretch
    # TODO: do the typography characters matter?
    # NOTE: | / and , are *not* sentence delimiters for my purpose


class WordTokenizerRe(RegexTokenizer):
    pattern = re.compile(r"""(?<=[.?!;:'"-])""")


class SentTokenizerRe(RegexTokenizer):
    pattern = re.compile(r"""(.*?[.?!;:])|(.+?$)""")


if not isinstance(nltk, ImportError):

    class WordTokenizerNLTK(Tokenizer):
        @classmethod
        @override
        def tokenize(cls, s: str) -> List[str]:
            return __word_tokenize_nltk(text=s, language=LANGUAGE)

    class SentTokenizerNLTK(Tokenizer):
        @classmethod
        @override
        def tokenize(cls, s: str) -> List[str]:
            return __sent_tokenize_nltk(text=s, language=LANGUAGE)



# STL
from typing import List, TypedDict

# PDM
import yaml
import pytest

# LOCAL
from sonatoki.Tokenizers import (
    SentTokenizerRe,
    WordTokenizerRe,
    SentTokenizerTok,
    WordTokenizerTok,
)

try:
    # PDM
    import nltk

    # LOCAL
    from sonatoki.Tokenizers import SentTokenizerNLTK, WordTokenizerNLTK

except ImportError as e:
    nltk = e
    # TODO: warn user


class TokenizerTest(TypedDict):
    name: str
    input: str
    output: List[str]
    should_be_equal: bool
    xfail: bool


def load_params_from_yaml(json_path: str) -> List[TokenizerTest]:
    with open(json_path) as f:
        return yaml.safe_load(f)


def load_tokenizer_tests(json_path: str) -> List[TokenizerTest]:
    tests = load_params_from_yaml(json_path)
    formatted_tests: List[TokenizerTest] = []
    for test in tests:
        formatted_tests.append(
            TokenizerTest(
                xfail=test.get("xfail", False),
                name=test.get("name", ""),
                input=test["input"],
                output=test.get("output", []),
                should_be_equal=test.get("should_be_equal", True),
            )
        )

    return formatted_tests


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_sentences.yml")
)
def test_sentences_re(test: TokenizerTest):
    if isinstance(nltk, ImportError):
        pytest.skip("nltk not installed")
    if test["xfail"]:
        pytest.xfail()

    transformed_re = SentTokenizerRe.tokenize(test["input"])
    transformed_nltk = SentTokenizerNLTK.tokenize(test["input"])
    assert (transformed_re == transformed_nltk) == test["should_be_equal"], test["name"]


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_words.yml")
)
def test_word_tokenize_re(test: TokenizerTest):
    if isinstance(nltk, ImportError):
        pytest.skip("nltk not installed")
    if test["xfail"]:
        pytest.xfail()

    transformed_re = WordTokenizerRe.tokenize(test["input"])
    transformed_nltk = WordTokenizerNLTK.tokenize(test["input"])
    assert (transformed_re == transformed_nltk) == test["should_be_equal"], test["name"]


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_sentences_tok.yml")
)
def test_sentences_tok(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    transformed_tok = SentTokenizerTok.tokenize(test["input"])
    if test["should_be_equal"]:
        assert transformed_tok == test["output"], test["name"]
    else:
        assert transformed_tok != test["output"], test["name"]


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_words_tok.yml")
)
def test_word_tokenize_tok(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    transformed_tok = WordTokenizerTok.tokenize(test["input"])
    if test["should_be_equal"]:
        assert transformed_tok == test["output"], test["name"]
    else:
        assert transformed_tok != test["output"], test["name"]



# STL
from typing import List, Type, Tuple

# LOCAL
from sonatoki.Filters import Filter
from sonatoki.Scorers import Number, Scorer
from sonatoki.Cleaners import Cleaner
from sonatoki.Tokenizers import Tokenizer
from sonatoki.Preprocessors import Preprocessor


class Ilo:
    __preprocessors: List[Type[Preprocessor]]
    __word_tokenizer: Type[Tokenizer]
    __cleaners: List[Type[Cleaner]]
    __ignoring_filters: List[Type[Filter]]
    __scoring_filters: List[Type[Filter]]
    __scorer: Type[Scorer]
    __passing_score: Number

    def __init__(
        self,
        preprocessors: List[Type[Preprocessor]],
        cleaners: List[Type[Cleaner]],
        ignoring_filters: List[Type[Filter]],
        scoring_filters: List[Type[Filter]],
        scorer: Type[Scorer],
        passing_score: Number,
        word_tokenizer: Type[Tokenizer],
    ):
        super().__init__()
        # avoid keeping a ref to user's list just in case
        self.__preprocessors = [*preprocessors]
        self.__word_tokenizer = word_tokenizer
        self.__cleaners = [*cleaners]
        self.__ignoring_filters = [*ignoring_filters]
        self.__scoring_filters = [*scoring_filters]
        self.__scorer = scorer
        self.__passing_score = passing_score

    def preprocess(self, msg: str) -> str:
        for p in self.__preprocessors:
            msg = p.process(msg)
        return msg

    def word_tokenize(self, msg: str) -> List[str]:
        """It is *highly* recommended that you run `ilo.preprocess` first."""
        return self.__word_tokenizer.tokenize(msg)

    def clean_token(self, token: str) -> str:
        for c in self.__cleaners:
            token = c.clean(token)
        return token

    def clean_tokens(self, tokens: List[str]) -> List[str]:
        # NOTE: tested, making a new list with a for loop *is* faster than:
        # list comp, generator comp, in-place replacement
        cleaned_tokens: List[str] = list()
        for token in tokens:
            cleaned_token = self.clean_token(token)
            if not cleaned_token:
                # TODO: warn user?
                continue
            cleaned_tokens.append(cleaned_token)
        return cleaned_tokens

    def _filter_token(self, token: str) -> bool:
        for f in self.__ignoring_filters:
            if f.filter(token):
                return True
        return False

    def filter_tokens(self, tokens: List[str]) -> List[str]:
        filtered_tokens: List[str] = []
        for token in tokens:
            if self._filter_token(token):
                continue
            # the ignoring filter is true if the token matches
            # the user wants to ignore these so keep non-matching tokens
            filtered_tokens.append(token)
        return filtered_tokens

    def score_tokens(self, tokens: List[str]) -> float:
        return self.__scorer.score(tokens, self.__scoring_filters)

    def _is_toki_pona(
        self, message: str
    ) -> Tuple[str, List[str], List[str], List[str], Number, bool]:
        """Returns all components of the processing algorithm:
        - Preprocessed message (str)
        - Tokenized message (list[str])
        - Filtered message (list[str])
        - Cleaned message (list[str])
        - Score (float)
        - Result (bool)"""
        preprocessed = self.preprocess(message)
        tokenized = self.word_tokenize(preprocessed)
        filtered = self.filter_tokens(tokenized)
        cleaned = self.clean_tokens(filtered)
        score = self.score_tokens(cleaned)
        result = score >= self.__passing_score

        return preprocessed, tokenized, filtered, cleaned, score, result

    def is_toki_pona(self, message: str) -> bool:
        """Determines whether a single statement is or is not Toki Pona."""
        *_, result = self._is_toki_pona(message)
        return result
